functor Gen (structure Seq : SEQUENCE) :>
sig
  (* generate b n s
   *
   * Requires 0 <= b <= 1. b is a nesting bias. Increasing b encourages the
   * resulting sequence to be more nested. At the extremes:
   *   b = 0 guarantees "()()()..."
   *   b = 1 guarantees "(((...)))"
   *
   * Requires n even and n >= 2. Produces seq of length n
   *
   * s is a seed. Any integer will do. *)
  val generate : real -> int -> int -> Paren.t Seq.t
end =
struct

  (* in case we need to tweak this to be faster! could change maxDepth to be
   * much smaller *)
  structure DotMix = MkDotMixParams (val init = 42
                                     val maxDepth = 500
                                     val mixNum = 4)

  structure AS = ArraySlice

  fun gcd (a, 0) = a
    | gcd (a, b) = gcd (b, a mod b)

  fun reduceRat (a, b) = let val d = gcd (a, b) in (a div d, b div d) end

  (* assume 0 <= b <= 1 *)
  fun approxRat n b =
    let
      fun approxRat' n (b : real) =
        if (n = 0) then (0,1) else
        let val {whole, frac} = Real.split (10.0 * b)
            val (x,y) = approxRat' (n-1) frac
        in ((Real.round whole)*y + x, 10*y)
        end
    in reduceRat (approxRat' n b)
    end

  (* r must be length n. generate sequence of parens and write them into r. *)
  fun generate (b : real) (n : int) (seed : int) =
    let
      val (x,y) = approxRat 3 b
      val biasPair = (x, y-x)
      fun writeGen r n s =
        if (n < 2) then raise Fail ("Cannot generate paren sequence of length " ^ Int.toString n) else
        if n = 2 then (AS.update (r, 0, Paren.L); AS.update (r, 1, Paren.R)) else
        if DotMix.biasedBool biasPair s then
          (AS.update (r, 0, Paren.L);
           AS.update (r, n-1, Paren.R);
           writeGen (AS.subslice (r, 1, (SOME (n-2)))) (n-2) (DotMix.next s))
        else
          let
            val (nl, nr) = if (n div 2) mod 2 = 1
                           then (n div 2 + 1, n div 2 - 1)
                           else (n div 2, n div 2)
            val (_, (sl, sr)) = DotMix.split s
          in Primitives.par (fn () => writeGen (AS.subslice (r, 0, SOME nl)) nl sl,
                             fn () => writeGen (AS.subslice (r, nl, NONE)) nr sr); ()
          end

      val r = AS.full (Array.array (n, Paren.L))
    in
      writeGen r n (DotMix.fromInt seed); Seq.tabulate (fn i => AS.sub (r, i)) n
    end

  (*fun generate (b : real) (n : int) (seed : int) =
    let
      val (x,y) = approxRat 3 b
      val biasPair = (x, y-x)
      fun gen n s =
        if (n < 2) then raise Fail (Int.toString n) else
        if n = 2 then Seq.fromList [Paren.L, Paren.R] else
        if DotMix.biasedBool biasPair s then
          Seq.append (Seq.$ Paren.L, Seq.append (gen (n-2) (DotMix.next s), Seq.$ Paren.R))
        else
          let
            val (nl, nr) = if (n div 2) mod 2 = 1
                           then (n div 2 + 1, n div 2 - 1)
                           else (n div 2, n div 2)
            val (_, (sl, sr)) = DotMix.split s
          in Seq.append (Primitives.par (fn () => gen nl sl,
                                         fn () => gen nr sr))
          end
    in
      gen n (DotMix.fromInt seed)
    end*)
end
